# import dis

# z = 10


# def foo(x):
#     def faf(y):
#         return x + y + z

#     return faf


# fn = foo(4)

# bytecode = dis.Bytecode(fn)
# print("=" * 80)
# for instr in bytecode:
#     print(instr)

# breakpoint()


from wrapt import ObjectProxy

from ovld import ovld, recurse


class AnnotatedProxy(ObjectProxy):
    def __init__(self, wrapped, **ann):
        super().__init__(wrapped)
        self._self_ann = ann

    @property
    def _(self):
        return self._self_ann


# @ovld(priority=10)
# def f(x: object):
#     return recurse.next(x)


@ovld
def f(xs: list):
    return [recurse(x) for x in xs]


@ovld
def f(xs: dict):
    return {recurse(k): recurse(v) for k, v in xs.items()}


@ovld
def f(x: str):
    return x


@ovld
def f(x: int):
    return x / 0


# # f({"a": [[[[1, 2, 3]]]]})
# # breakpoint()
# res = f({"a": [[[[1, 2, 3]]]]})
# print(res)

# data = {"a": 2}
# print(f(AnnotatedProxy(data, x=4)))



import inspect
from types import FunctionType

recurse = object()


class Conformer:
    __slots__ = ("code", "orig_fn", "renamed_fn", "ovld", "code2")

    def __init__(self, ovld, orig_fn, renamed_fn):
        self.ovld = ovld
        self.orig_fn = orig_fn
        self.renamed_fn = renamed_fn
        self.code = orig_fn.__code__

    def __conform__(self, new):
        if isinstance(new, FunctionType):
            new_fn = new
            new_code = new.__code__
        else:
            new_fn = None
            new_code = new

        if new_code is None:
            self.ovld.unregister(self.orig_fn)

        elif new_fn is None:  # pragma: no cover
            # Not entirely sure if this ever happens
            self.renamed_fn.__code__ = new_code

        elif inspect.signature(self.orig_fn) != inspect.signature(new_fn):
            self.ovld.unregister(self.orig_fn)
            self.ovld.register(new_fn)

        else:
            self.renamed_fn.__code__ = rename_code(
                new_code, self.renamed_fn.__code__.co_name
            )

        from codefind import code_registry

        code_registry.update_cache_entry(self, self.code, new_code)

        self.code = new_code


def rename_code(co, newname):  # pragma: no cover
    if hasattr(co, "replace"):
        if hasattr(co, "co_qualname"):
            return co.replace(co_name=newname, co_qualname=newname)
        else:
            return co.replace(co_name=newname)
    else:
        return type(co)(
            co.co_argcount,
            co.co_kwonlyargcount,
            co.co_nlocals,
            co.co_stacksize,
            co.co_flags,
            co.co_code,
            co.co_consts,
            co.co_names,
            co.co_varnames,
            co.co_filename,
            newname,
            co.co_firstlineno,
            co.co_lnotab,
            co.co_freevars,
            co.co_cellvars,
        )


def rename_function(fn, newname):
    """Create a copy of the function with a different name."""
    newcode = rename_code(fn.__code__, newname)
    new_fn = FunctionType(
        newcode, fn.__globals__, newname, fn.__defaults__, fn.__closure__
    )
    new_fn.__annotations__ = fn.__annotations__
    return new_fn



from .core import (
    MultiTypeMap,
    Ovld,
    OvldBase,
    OvldCall,
    OvldMC,
    TypeMap,
    extend_super,
    is_ovld,
    ovld,
    ovld_dispatch,
)
from .utils import (
    BOOTSTRAP,
    MISSING,
    Dataclass,
    Named,
    deferred,
    exactly,
    has_attribute,
    keyword_decorator,
    meta,
    strict_subclass,
)
from .version import version as __version__

__all__ = [
    "MultiTypeMap",
    "Ovld",
    "OvldBase",
    "OvldCall",
    "OvldMC",
    "TypeMap",
    "extend_super",
    "is_ovld",
    "ovld",
    "ovld_dispatch",
    "BOOTSTRAP",
    "MISSING",
    "Dataclass",
    "Named",
    "deferred",
    "exactly",
    "has_attribute",
    "meta",
    "keyword_decorator",
    "strict_subclass",
]
